from typing import Optional, List, Any
from pydantic import BaseModel, ValidationError, field_validator, Field
import json


class Priority(BaseModel):
    priority: Any
    color: str


class Status(BaseModel):
    status: str | None = None
    color: str | None = None
    hide_label: bool | None = None


class StatusElement(BaseModel):
    id: Optional[str]
    status: str

    orderindex: int
    color: str

    type: str


class Asssignee(BaseModel):
    id: str = None
    color: str = None
    username: str = None
    initials: str = None

    profilePicture: str = None


class ListFolder(BaseModel):
    id: str
    name: str

    hidden: Optional[bool]

    access: bool


class Assignee(BaseModel):
    id: Optional[str] = None
    username: Optional[str] = None
    email: Optional[str] = None
    color: Optional[str] = None
    initials: Optional[str] = None
    profilePicture: Optional[str] = None


class SingleList(BaseModel):
    id: str
    name: str
    content: Optional[str] = None
    status: Optional[dict] = None
    priority: Optional[dict] = None
    assignee: Optional[Assignee] = None
    task_count: Optional[int] = 0

    @classmethod
    def build_list(cls, data: dict):
        if data.get('assignee') is None:
            data['assignee'] = Assignee()
        return cls(**data)


class AllLists(BaseModel):
    lists: List[SingleList] = None

    # return a list of lists

    def build_lists(self):
        return AllLists(**self)


class ChecklistItem(BaseModel):
    id: str = None
    name: str = None

    orderindex: int = None

    assignee: Optional[Asssignee]


class Checklist(BaseModel):
    id: Optional[str]

    task_id: str = None
    name: str = None

    orderindex: int = None

    resolved: int = None

    unresolved: int = None

    items: List[ChecklistItem] = None

    def add_item(self, client_instance, name: str, assignee: str = None):
        return client_instance.create_checklist_item(
            self.id, name=name, assignee=assignee
        )


class Checklists(BaseModel):
    checklist: Checklist

    def build_checklist(self):
        final_checklist = Checklists(**self)

        return final_checklist.checklist


class Attachment(BaseModel):
    id: str

    version: int
    date: str
    title: str

    extension: str

    thumbnail_small: str

    thumbnail_large: str
    url: str

    def build_attachment(self):
        return Attachment(**self)


class User(BaseModel):
    id: Optional[str | int] = None
    username: Optional[str] = None
    initials: Optional[str] = None
    email: Optional[str] = None
    color: Optional[str] = None
    profilePicture: Optional[str] = None
    role: Optional[int] = None
    custom_role: Optional[None] = None
    last_active: Optional[str] = None
    date_joined: Optional[str] = None
    date_invited: Optional[str] = None

    @field_validator('id')
    def convert_id_to_str(cls, v):
        return str(v) if v is not None else None


class AssignedBy(BaseModel):
    id: str = None
    username: str = None
    initials: str = None
    email: str = None
    color: str = None
    profile_picture: str = None


class CommentComment(BaseModel):
    text: str = None


class Comment(BaseModel):
    id: str = None

    comment: List[CommentComment] = None

    comment_text: str = None

    user: AssignedBy = None

    resolved: bool = None

    assignee: AssignedBy = None

    assigned_by: AssignedBy = None

    reactions: List[Any] = None
    date: str = None
    hist_id: str = None

    def build_comment(self):
        return Comment(**self)


class Comments(BaseModel):
    comments: List[Comment] = None

    def __iter__(self):
        return iter(self.comments)

    def build_comments(self):
        return Comments(**self)


class Creator(BaseModel):
    id: int = None
    username: str = None
    color: str = None
    profile_picture: str = None


class Option(BaseModel):
    id: Optional[str]

    name: Optional[str]

    color: Optional[str]

    order_index: Optional[int]


class TypeConfig(BaseModel):
    default: Optional[int]

    placeholder: Optional[str]

    new_drop_down: Optional[bool]

    options: Optional[List[Option]]

    include_guests: Optional[bool]

    include_team_members: Optional[bool]


class CustomItems:
    enabled: bool = None


class DueDates(BaseModel):
    enabled: bool = None

    start_date: bool = None

    remap_due_dates: bool = None

    remap_closed_due_date: bool = None


class CustomField(BaseModel):
    id: str = None
    name: str = None

    type: str = None

    type_config: TypeConfig = None
    date_created: str = None

    hide_from_guests: bool = None

    value: Optional[Any] = None

    required: bool = None


class TimeTracking(BaseModel):
    enabled: bool = False

    harvest: bool = False

    rollup: bool = False


class Sprints(BaseModel):
    enabled: bool = False


class Points(BaseModel):
    enabled: bool = False


class Zoom(BaseModel):
    enabled: bool = False


class Milestones(BaseModel):
    enabled: bool = False


class Emails(BaseModel):
    enabled: bool = False


class CustomItems(BaseModel):
    enabled: bool = False


class MultipleAssignees(BaseModel):
    enabled: bool = False


class TagsStatus(BaseModel):
    enabled: bool = False


class CustomFieldsStatus(BaseModel):
    enabled: bool = False


class DependencyWarning(BaseModel):
    enabled: bool = False


class TimeEstimateStatus(BaseModel):
    enabled: bool = False


class RemapDependenciesStatus(BaseModel):
    enabled: bool = False


class ChecklistsStatus(BaseModel):
    enabled: bool = False


class PortfoliosStatus(BaseModel):
    enabled: bool = False


class Features(BaseModel):
    due_dates: DueDates | None = None
    multiple_assignees: MultipleAssignees | None = None
    sprints: Sprints | None = None
    start_date: bool = False
    remap_due_dates: bool = False
    remap_closed_due_date: bool = False
    time_tracking: Optional[TimeTracking] = None
    tags: Optional[TagsStatus] = None
    time_estimates: Optional[TimeEstimateStatus] = None
    checklists: Optional[ChecklistsStatus] = None
    custom_fields: Optional[CustomFieldsStatus] = None
    remap_dependencies: Optional[RemapDependenciesStatus] = None
    dependency_warning: DependencyWarning | None = None
    portfolios: Optional[PortfoliosStatus] = None
    points: Points | None = None
    custom_items: CustomItems | None = None
    zoom: Zoom | None = None
    milestones: Milestones | None = None
    emails: Emails | None = None

    model_config = {
        "validate_assignment": True
    }

    @field_validator("time_tracking", mode="before")
    def set_tt(cls, time_tracking):
        return time_tracking or {"enabled": False}

    @field_validator("custom_fields", mode="before")
    def set_cf(cls, custom_fields):
        return custom_fields or {"enabled": False}

    @field_validator("tags", mode="before")
    def set_tags(cls, tags):
        return tags or {"enabled": False}

    @field_validator("multiple_assignees", mode="before")
    def set_ma(cls, multiple_assignees):
        return multiple_assignees or {"enabled": False}

    @field_validator("checklists", mode="before")
    def set_checklists(cls, checklists):
        return checklists or {"enabled": False}

    @field_validator("portfolios", mode="before")
    def set_portfolios(cls, portfolios):
        return portfolios or {"enabled": False}


class SpaceFeatures(BaseModel):
    due_dates: bool = False

    multiple_assignees: bool = False

    start_date: bool = False

    remap_due_dates: bool = False

    remap_closed_due_date: bool = False

    time_tracking: bool = False

    tags: bool = False

    time_estimates: bool = False

    checklists: bool = False

    custom_fields: bool = False

    remap_dependencies: bool = False

    dependency_warning: bool = False

    portfolios: bool = False

    points: bool = False

    custom_items: bool = False

    zoom: bool = False

    milestones: bool = False

    emails: bool = False

    @property
    def all_features(self):
        return {
            "due_dates": {
                "enabled": self.due_dates,
                "start_date": self.start_date,
                "remap_due_dates": self.remap_due_dates,
                "remap_closed_due_date": self.remap_closed_due_date,
            },
            "time_tracking": {"enabled": self.time_tracking},
            "tags": {"enabled": self.tags},
            "time_estimates": {"enabled": self.time_estimates},
            "checklists": {"enabled": self.checklists},
            "custom_fields": {"enabled": self.custom_fields},
            "remap_dependencies": {"enabled": self.remap_dependencies},
            "dependency_warning": {"enabled": self.dependency_warning},
            "portfolios": {"enabled": self.portfolios},
            "milestones": {"enabled": self.milestones},
        }


class Space(BaseModel):
    id: Optional[str] = None

    name: Optional[str] = None

    access: Optional[bool] = None

    features: Optional[Features]

    multiple_assignees: Optional[bool] = None

    private: Optional[bool] = False

    statuses: Optional[List[Status]] = None

    archived: Optional[bool] = None

    def build_space(self):
        return Space(**self)


class Spaces(BaseModel):
    spaces: List[Space] = None

    def __iter__(self):
        return iter(self.spaces)

    def build_spaces(self):
        return Spaces(**self)


class Folder(BaseModel):
    id: str = None
    name: str = None

    orderindex: int = None

    override_statuses: bool = False

    hidden: bool = False

    space: Optional[dict] = None

    task_count: int = None

    lists: List[SingleList] = []

    def build_folder(self):
        if self.space:
            self.space = Space(**self.space)
        return self

    def delete(self, client_instance):
        model = "folder/"

        deleted_folder_status = client_instance._delete_request(model, self.id)


class Folders(BaseModel):
    folders: List[Folder] = None

    def build_folders(self):
        return Folders(**self)


class Priority(BaseModel):
    id: int = None

    priority: Any = None
    color: str = None

    orderindex: str = None


class Status(BaseModel):
    id: Optional[str] = None
    status: str = None
    color: str = None

    orderindex: int = None

    type: str = None


class ClickupList(BaseModel):
    id: str = None


# class Folder(BaseModel):

#     id: str = None


class Task(BaseModel):
    id: Optional[str] = None
    custom_id: Optional[str] = None
    name: Optional[str] = None

    text_content: Optional[str] = None
    description: Optional[str] = None

    status: Optional[Status] = None

    orderindex: Optional[str] = None
    date_created: Optional[str] = None
    date_updated: Optional[str] = None
    date_closed: Optional[str] = None

    creator: Optional[Creator] = None

    assignees: Optional[List[Asssignee]] = None

    task_checklists: Optional[List[Any]] = Field(None, alias="checklists")

    task_tags: Optional[List[Any]] = Field(None, alias="tags")
    parent: Optional[str] = None

    priority: Optional[Any] = None
    due_date: Optional[str] = None
    start_date: Optional[str] = None
    time_estimate: Optional[str] = None

    time_spent: Optional[str] = None

    custom_fields: Optional[List[CustomField]] = None
    list: Optional[ClickupList] = None

    folder: Optional[Folder] = None

    space: Optional[Folder] = None
    url: Optional[str] = ""

    def build_task(self):
        return Task(**self)

    def delete(self):
        client.ClickUpClient.delete_task(self, self.id)

    def upload_attachment(self, client_instance, file_path: str):
        return client_instance.upload_attachment(self.id, file_path)

    def update(
        self,
        client_instance,
        name: str = None,
        description: str = None,
        status: str = None,
        priority: Any = None,
        time_estimate: int = None,
        archived: bool = None,
        add_assignees: List[str] = None,
        remove_assignees: List[int] = None,
    ):
        return client_instance.update_task(
            self.id,
            name,
            description,
            status,
            priority,
            time_estimate,
            archived,
            add_assignees,
            remove_assignees,
        )

    def add_comment(
        self,
        client_instance,
        comment_text: str,
        assignee: str = None,
        notify_all: bool = True,
    ):
        return client_instance.create_task_comment(
            self.id, comment_text, assignee, notify_all
        )

    def get_comments(self, client_instance):
        return client_instance.get_task_comments(self.id)


class Tasks(BaseModel):
    tasks: List[Task] = None

    def __iter__(self):
        return iter(self.tasks)

    def build_tasks(self):
        return Tasks(**self)


class InvitedBy(BaseModel):
    id: str = None
    username: str = None
    color: str = None
    email: str = None
    initials: str = None
    profile_picture: None = None


class Member(BaseModel):
    user: User

    invited_by: Optional[InvitedBy] = None


class Members(BaseModel):
    members: List[User] = None

    def __iter__(self):
        return iter(self.members)

    def build_members(self):
        return Members(**self)


class Team(BaseModel):
    id: Optional[str] = None
    name: Optional[str] = None
    color: Optional[str] = None
    avatar: Optional[str] = None
    members: Optional[List[Member]] = None


class Teams(BaseModel):
    teams: List[Team] = None

    def __iter__(self):
        return iter(self.teams)

    def build_teams(self):
        return Teams(**self)


class Goal(BaseModel):
    id: Optional[str] = None
    name: Optional[str] = None
    team_id: Optional[int] = None
    date_created: Optional[str] = None
    start_date: Optional[str] = None
    due_date: Optional[str] = None
    description: Optional[str] = None
    private: Optional[bool] = None
    archived: Optional[bool] = None
    creator: Optional[int] = None
    color: Optional[str] = None
    pretty_id: Optional[int] = None
    multiple_owners: Optional[bool] = None
    folder_id: Optional[str] = None
    members: Optional[List[User]] = None
    owners: Optional[List[User]] = None
    key_results: Optional[List[Any]] = None
    percent_completed: Optional[int] = None
    history: Optional[List[Any]] = None
    pretty_url: Optional[str] = None

    def build_goal(self):
        return Goal(**self)


class Goals(BaseModel):
    goal: Goal

    def build_goals(self):
        built_goal = Goals(**self)

        return built_goal.goal


class GoalsList(BaseModel):
    goals: List[Goal] = None
    folders: List[Folder] = None

    def __iter__(self):
        return iter(self.goals)

    def build_goals(self):
        return GoalsList(**self)


class Tag(BaseModel):
    name: str = None

    tag_fg: str = None

    tag_bg: str = None

    def build_tag(self):
        return Tag(**self)


class Tags(BaseModel):
    tags: List[Tag] = None

    def __iter__(self):
        return iter(self.tags)

    def build_tags(self):
        return Tags(**self)


class Shared(BaseModel):
    tasks: Optional[List[Tasks]]

    lists: Optional[List[SingleList]]

    folders: Optional[List[Folder]]

    def build_shared(self):
        return Shared(**self)

    def __iter__(self):
        return iter(self.shared)


class SharedHierarchy(BaseModel):
    shared: Shared

    def build_shared(self):
        return SharedHierarchy(**self)

    def __iter__(self):
        return iter(self.shared)


class TimeTrackingData(BaseModel):
    id: str = ""
    task: Task = None
    wid: str = ""
    user: User = None
    billable: bool = False
    start: str = ""
    end: str = ""
    duration: int = None
    description: str = ""
    tags: List[Tag] = None
    source: str = ""
    at: str = ""

    def build_data(self):
        return TimeTrackingData(**self)


class TimeTrackingDataList(BaseModel):
    data: List[TimeTrackingData] = None

    def build_data(self):
        return TimeTrackingDataList(**self)

    def __iter__(self):
        return iter(self.data)


class TimeTrackingDataSingle(BaseModel):
    data: TimeTrackingData = None

    def build_data(self):
        return TimeTrackingDataSingle(**self)

    def __iter__(self):
        return iter(self.data)
